from django_htmx_ui.utils import ContextProperty
from django_htmx_ui.views.mixins import ResponseNoContentMixin, FormMixin, InstanceMixin


class CrudMixin:

    @property
    def permission(self):
        return True

    def on_get(self, *args, **kwargs):
        if not self.permission:
            raise ValueError('Pemission error')
        return super().on_get(*args, **kwargs)

    def on_post(self, *args, **kwargs):
        if not self.permission:
            raise ValueError('Pemission error')
        return super().on_post(*args, **kwargs)


class CrudCreateMixin(CrudMixin, FormMixin):
    pass


class CrudRetrieveMixin(CrudMixin):
    filter = {}

    def filters_get(self):
        return {
            key[7:]: value
            for key, value in self.request.GET.items()
            if key.startswith('filter_')
        }

    @ContextProperty
    def instances(self):
        return self.module.MODEL.objects.filter(**self.filter).filter(**self.filters_get())


class CrudListMixin(CrudRetrieveMixin):
    pass


class CrudUpdateMixin(InstanceMixin, CrudMixin):
    pass


class CrudDisplayMixin(InstanceMixin, CrudMixin):
    pass


class CrudActionMixin(ResponseNoContentMixin, InstanceMixin, CrudMixin):
    pass


class CrudDeleteMixin(CrudActionMixin):

    def on_post(self, request, *args, **kwargs):
        count, deleted = self.instance.delete()
        if count:
            self.on_post_success_message(request, *args, **kwargs)
            return self.on_post_success(request, *args, **kwargs)
        else:
            self.on_post_invalid_message(request, *args, **kwargs)
            return self.on_post_invalid(request, *args, **kwargs)

    def on_post_success_message(self, request, *args, **kwargs):
        self.message_success('%s deleted!' % self.instance)

    def on_post_invalid_message(self, request, *args, **kwargs):
        self.message_error('%s not deleted!' % self.instance)



from django.core.exceptions import ViewDoesNotExist
from django.shortcuts import redirect

from django_htmx_ui.utils import ContextProperty, ContextCachedProperty, UrlView, to_snake_case


class OriginTemplateMixin:

    def get(self, request, *args, **kwargs):
        if request.htmx and OriginTemplateMixin in self.__class__.__bases__:
            raise ViewDoesNotExist("View '%s' is an 'OriginTemplateMixin'." % self.__class__.__name__)
        else:
            return super().get(request, *args, **kwargs)


class PartialTemplateMixin:
    redirect_partial = '/'

    def get(self, request, *args, **kwargs):
        if request.htmx:
            return super().get(request, *args, **kwargs)
        elif self.redirect_partial:
            return redirect(self.redirect_partial)
        else:
            raise ViewDoesNotExist("View '%s' is a 'PartialTemplateMixin'." % self.__class__.__name__)


class FormMixin:
    form_initial = {}

    @property
    def form_instance(self):
        return getattr(self, 'instance', None)

    @ContextCachedProperty
    def form(self):
        Form = getattr(self.__class__, 'Form', None)
        if Form:
            if self.request.method == 'GET':
                if self.form_instance:
                    return Form(instance=self.form_instance, initial=self.form_initial)
                else:
                    return Form(initial=self.form_initial)
            if self.request.method == 'POST':
                if self.form_instance:
                    return Form(self.request.POST, instance=self.form_instance)
                else:
                    return Form(self.request.POST)

    def on_post(self, request, *args, **kwargs):
        if self.form.is_valid():
            self.form.save()
            self.on_post_success_message(request, *args, **kwargs)
            return self.on_post_success(request, *args, **kwargs)
        else:
            self.on_post_invalid_message(request, *args, **kwargs)
            return self.on_post_invalid(request, *args, **kwargs)

    def on_post_success(self, request, *args, **kwargs):
        pass

    def on_post_success_message(self, request, *args, **kwargs):
        self.message_success('Saved!')

    def on_post_invalid(self, request, *args, **kwargs):
        pass

    def on_post_invalid_message(self, request, *args, **kwargs):
        self.message_error('Not saved!')


class InstanceMixin(FormMixin):

    @classmethod
    @property
    def path_route(cls):
        return '(?P<pk>\w+)/' + super().path_route

    @ContextProperty
    def url(self):
        return UrlView(self, self.instance.pk)

    @ContextCachedProperty
    def instance(self):
        return self.module.MODEL.objects.get(pk=self.request.resolver_match.kwargs['pk'])

    @property
    def instance_slug(self):
        return '%s_%s' % (to_snake_case(self.__class__.__name__), self.instance.pk)

    @ContextProperty
    def title(self):
        return str(self.instance)

    @property
    def permission(self):
        return self.instance

    def on_post_success_message(self, request, *args, **kwargs):
        self.message_success('%s saved!' % self.instance)

    def on_post_invalid_message(self, request, *args, **kwargs):
        self.message_error('%s not saved!' % self.instance)


class ResponseNoContentMixin:

    def get(self, request, *args, **kwargs):
        self.response_no_content()
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.response_no_content()
        return super().post(request, *args, **kwargs)


class TabsMixin:

    class Tabs:

        class Link:
            index = None

            def __init__(self, title, url, slug=None):
                self.title = title
                self.url = url
                self._slug = slug

            @property
            def slug(self):
                return self._slug or self.index

        def __init__(self, *links, selected=0, remember=False):
            self.selected = selected
            self.remember = remember
            self.links = links

            for i, l in enumerate(self.links):
                l.index = i

        @property
        def active(self):
            return self.links[self.selected]

    def __init__(self, *args, **kwargs):
        class TabsView(TabsMixin.Tabs):
            view = self
        self.Tabs = TabsView
        super().__init__(*args, **kwargs)

    @ContextProperty
    def tabs(self):
        raise NotImplementedError()

    @classmethod
    @property
    def slug_tab(cls):
        return f'{cls.slug_module}_tab'

    @property
    def tab_query_var(self):
        return self.slug_tab

    @classmethod
    @property
    def path_route(cls):
        return super().path_route + f'(?:(?P<{cls.slug_tab}>\w+)/)?'

    def on_get(self, request, *args, **kwargs):
        super().on_get(request, *args, **kwargs)

        session_key = f'tabs-remember-{self.slug_tab}'
        if self.tabs.remember and session_key in request.session:
            self.tabs.selected = int(request.session[session_key])

        if self.tab_query_var in request.GET:
            self.tabs.selected = int(request.GET[self.tab_query_var])

        slug_tab_value = self.request.resolver_match.kwargs.get(self.slug_tab)
        if slug_tab_value:
            for link in self.tabs.links:
                if link.slug == slug_tab_value:
                    self.tabs.selected = int(link.index)
                    break
            else:
                raise ValueError(f"Tab slug '{slug_tab_value}' not found.")

        if self.tabs.active.slug:
            new_path = self.url(**{self.slug_tab: self.tabs.active.slug})
            slug_location_bar = self.location_bar.resolver_match.kwargs.get(self.slug_tab)
            push = slug_location_bar not in (None, slug_tab_value)
            self.location_bar(path=new_path, push=push)

        if self.tabs.remember:
            request.session[session_key] = int(self.tabs.selected)


class ModalMixin:

    class Modal:

        def __init__(self, url, _id=None):
            self.url = url
            self.id = _id or f'modal_{self.view.slug_global}'

    def __init__(self, *args, **kwargs):
        class ModalView(ModalMixin.Modal):
            view = self
        self.Modal = ModalView
        super().__init__(*args, **kwargs)




import importlib
import inspect
import os

from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponse
from django.shortcuts import redirect
from django.template import engines
from django.urls import re_path
from django.views.generic import TemplateView, RedirectView
from django_htmx.http import HttpResponseLocation, trigger_client_event, HttpResponseClientRedirect

from django_htmx_ui.utils import ContextProperty, ContextCachedProperty, merge, to_snake_case, UrlView, Location
from django_htmx_ui.views.mixins import OriginTemplateMixin


class BaseTemplateView(TemplateView):
    response = None

    def setup(self, request, *args, **kwargs):
        self.headers = {}
        self.triggers = []
        self.request = request
        self.location_bar = Location.create_from_url(self.bar_url())
        self.location_req = Location.create_from_url(request.META['PATH_INFO'])
        self.add_context('request', request)
        return super().setup(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        ret = self.on_get(request, *args, **kwargs)
        if ret:
            return ret
        elif self.response:
            return self.response_prepare(self.response)
        else:
            return self.response_prepare(super().get(request, *args, **kwargs))

    def on_get(self, request, *args, **kwargs):
        pass

    def post(self, request, *args, **kwargs):
        ret = self.on_post(request, *args, **kwargs)
        if ret:
            return ret
        elif self.response:
            return self.response_prepare(self.response)
        else:
            return self.response_prepare(super().get(request, *args, **kwargs))

    def on_post(self, request, *args, **kwargs):
        pass

    def render_to_response(self, context, **response_kwargs):
        response = super().render_to_response(context, **response_kwargs)
        return self.response_prepare(response)

    def response_location(self, *args, **kwargs):
        self.response = HttpResponseLocation(*args, **kwargs)
        return self.response

    def response_no_content(self):
        self.response = HttpResponse(status=204)
        return self.response

    def response_prepare(self, response):
        for func in (self.apply_triggers, self.apply_headers, self.apply_location):
            response = func(response)
        return response

    def apply_headers(self, response):
        for key, value in self.headers.items():
            response[key] = value
        return response

    def apply_triggers(self, response):
        for args, kwargs in self.triggers:
            trigger_client_event(response, *args, **kwargs)
        return response

    def apply_location(self, response):
        if self.request.htmx and self.location_bar != Location.create_from_url(self.bar_url()):
            response['HX-Push-Url' if self.location_bar.push else 'HX-Replace-Url'] = str(self.location_bar)
        return response

    def trigger_client_event(self, *args, **kwargs):
        self.triggers.append((args, kwargs))

    def decorators_context(self):
        return {
            name: getattr(self, name)
            for name, method in inspect.getmembers_static(
                self.__class__,
                lambda o:
                    isinstance(o, (
                        ContextProperty,
                        ContextCachedProperty,
                    ))
            )
        }

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            **self.decorators_context(),
            **getattr(self, '_context', {}),
        }

    def get_template_names(self):
        if not self.request.htmx:
            return self.template_origin
        else:
            return super().get_template_names()

    def add_context(self, key, value):
        setattr(self, '_context', merge(getattr(self, '_context', {}), {key: value}))

    @classmethod
    @property
    def module(cls):
        return importlib.import_module(cls.__module__)

    @classmethod
    @property
    def slug(cls):
        return to_snake_case(cls.__name__)

    @classmethod
    @property
    def instance_slug(cls):
        return cls.slug

    @classmethod
    @property
    def slug_module(cls):
        slug = getattr(cls.module, 'SLUG', cls.__module__.replace('.', '_'))
        return slug

    @classmethod
    @property
    def slug_global(cls):
        return f'{cls.slug_module}_{cls.slug}'

    @classmethod
    @property
    def path_route(cls):
        return cls.slug + '/'

    @classmethod
    @property
    def path(cls):
        return re_path(rf'^{cls.path_route}$', cls.as_view(), name=cls.slug)

    def redirect(self, url):
        if self.request.htmx:
            return HttpResponseClientRedirect(url)
        else:
            return redirect(url)

    @ContextProperty
    def url(self):
        return UrlView(self)

    def bar_url(self):
        return self.request.headers.get('HX-Current-URL', self.request.META['PATH_INFO'])

    @classmethod
    @property
    def templates_dir(cls):
        if hasattr(cls.module, 'TEMPLATES_DIR'):
            return cls.module.TEMPLATES_DIR
        else:
            app, views, crud = cls.__module__.split('.')
            return f'{app}/{crud}/'

    @classmethod
    @property
    def template_file(cls):
        return cls.slug + '.html'

    @classmethod
    @property
    def template_name(cls):
        return cls.templates_dir + cls.template_file

    @classmethod
    @property
    def template_origin(cls):
        for super_cls in cls.__mro__:
            if OriginTemplateMixin in super_cls.__bases__:
                return super_cls.template_name
        raise ValueError('You must define an `OriginTemplateMixin`.')

    def render(self, context):
        template = engines['django'].get_template(self.template_name)
        return template(self.get_context_data(**context))

    @ContextProperty
    def slug_panel(self):
        return self.slug_module

    @ContextProperty
    def title(self):
        return getattr(self.module, 'TITLE', self.project_title)

    @ContextProperty
    def icon(self):
        return getattr(self.module, 'ICON', '')

    @ContextProperty
    def project_title(self):
        return os.environ['DJANGO_SETTINGS_MODULE'].split('.')[0]

    def message_info(self, message):
        messages.info(self.request, message)

    def message_success(self, message):
        messages.success(self.request, message)

    def message_warning(self, message):
        messages.warning(self.request, message)

    def message_error(self, message):
        messages.error(self.request, message)


class PublicTemplateView(BaseTemplateView):
    pass


class PrivateTemplateView(LoginRequiredMixin, PublicTemplateView):

    @ContextProperty
    def user(self):
        return self.request.user



from django.templatetags.static import static
from django.urls import reverse
from django.utils import timezone
from django_htmx.jinja import django_htmx_script
from django.utils.translation import gettext, ngettext
import humanize

from jinja2 import Environment


def environment(**options):
    options.update({'extensions':['jinja2.ext.i18n']})
    env = Environment(**options)
    env.install_gettext_callables(gettext=gettext, ngettext=ngettext, newstyle=True)
    env.globals.update({
        'static': static,
        'url': reverse,
        'humanize': humanize,
        'now': timezone.now,
        'django_htmx_script': django_htmx_script,
    })
    return env



import functools
import importlib
import inspect
import re
from urllib.parse import urlencode, urlparse, parse_qsl

from django.urls import path, include, reverse, resolve
from django.shortcuts import redirect
from django_htmx.http import HttpResponseClientRedirect


class ContextProperty(property):
    pass


class ContextCachedProperty(functools.cached_property):
    pass


class Url:

    class Query:

        def __init__(self, query_list, url):
            self.query_list = query_list
            self.url = url

        def __call__(self, *args, **kwargs):
            self.add(*args, **kwargs)
            return self.url

        def reset(self, *args, **kwargs):
            self.query_list = []
            self.add(*args, **kwargs)
            return self.url

        def set(self, query_list):
            self.query_list = query_list
            return self.url

        def add(self, *args, **kwargs):
            for name, value in args:
                self.query_list.append((name, value))
            self.update(**kwargs)
            return self.url

        def remove(self, name):
            self.query_list = [(n, v) for n, v in self.query_list if n != name]
            return self.url

        def update(self, **kwargs):
            for name, value in kwargs.items():
                self.remove(name)
                self.query_list.append((name, value))
            return self.url

    def __init__(self, path, query_list):
        self.path = str(path)
        self.query = Url.Query(query_list, self)

    def __call__(self, path=None, query_list=None):
        if path is not None:
            self.path = path
        if query_list is not None:
            self.query = Url.Query(query_list, self)
        return self

    def __str__(self):
        urlencoded = urlencode(self.query.query_list)
        return (str(self.path) + '?' + urlencoded) if urlencoded else str(self.path)

    def __eq__(self, other):
        return str(self) == str(other)

    @property
    def resolver_match(self):
        return resolve(self.path)

    @property
    def view(self):
        return self.resolver_match.func.view_class

    @classmethod
    def create(cls, view_ref, *args, **kwargs):
        if type(view_ref) is str and view_ref.find('.') == -1:
            path = reverse(view_ref, args=args, kwargs=kwargs)
        else:
            if type(view_ref) is str and view_ref.find('.') >= 0:
                module_name, class_name = view_ref.rsplit(".", 1)
                module = importlib.import_module(module_name)
                klass = getattr(module, class_name)
            elif type(view_ref) is type:
                klass = view_ref
            else:
                klass = view_ref.__class__
            app, views, crud = klass.__module__.split('.')
            path = reverse(
                f'{app}:{crud}:{klass.slug}',
                **dict(args=args) if args else dict(kwargs=kwargs)
            )

        url = cls(path, [])
        return url


class UrlView:

    def __init__(self, view_ref, *args, **kwargs):
        self.view_ref = view_ref
        self.args = args
        self.kwargs = kwargs

    def update(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        del self._url
        return self

    @functools.cached_property
    def _url(self):
        return Url.create(self.view_ref, *self.args, **self.kwargs)

    @property
    def path(self):
        return self._url.path

    @property
    def query(self):
        return self._url.query

    @property
    def resolver_match(self):
        return self._url.resolver_match

    @property
    def view(self):
        return self._url.view

    def __str__(self):
        return str(self._url)

    def __call__(self, view_ref=None, *args, **kwargs):
        if view_ref is None:
            return UrlView(self.view_ref, *(args or self.args), **{**self.kwargs, **kwargs})
        else:
            return UrlView(view_ref, *args, **kwargs)


class Location(Url):

    def __init__(self, path, query_list, push=False):
        self.push = push
        super().__init__(path, query_list)

    def __call__(self, path=None, query_list=None, push=False):
        if push is not None:
            self.push = push
        return super().__call__(path, query_list)

    @property
    def resolver_match(self):
        return resolve(self.path)

    @classmethod
    def create_from_url(cls, location_url):
        parsed_url = urlparse(location_url)
        parsed_path = parsed_url.path
        parsed_qsl = parse_qsl(parsed_url.query)

        url = cls(parsed_path, parsed_qsl)
        return url


def collect_paths(module, app_name):
    from django_htmx_ui.views.generic import BaseTemplateView
    from django_htmx_ui.views.mixins import OriginTemplateMixin
    members = inspect.getmembers(
        module,
        lambda o:
            inspect.isclass(o) and
            issubclass(o, BaseTemplateView) and
            OriginTemplateMixin not in o.__bases__ and
            o.__module__ == module.__name__
    )
    slug = getattr(module, 'SLUG', module.__name__.split('.')[-1])
    path_route = getattr(module, 'PATH_ROOT', slug + '/')
    includes = [
        klass.path
        for name, klass in members
        if hasattr(klass, 'path')
    ]
    paths = path(path_route, include((includes, app_name), namespace=slug))
    return paths


def x_redirect(request,  url):
    if request.htmx:
        return HttpResponseClientRedirect(url)
    else:
        return redirect(url)


def paginated(items_per_page):
    def inner(func):
        def wrapper(self, *args, **kwargs):
            page = max(1, int(self.request.GET.get('page', 1)))
            index = (page - 1) * items_per_page
            return func(self, *args, **kwargs)[index:index+items_per_page]
        return wrapper
    return inner


def indexed(field, items_per_round, reverse=False):
    def inner(func):
        def wrapper(self, *args, **kwargs):
            qs = func(self, *args, **kwargs).order_by(f'-{field}' if reverse else field)
            last_field = getattr(self, f'last_{field}', self.request.GET.get(f'last_{field}'))
            if last_field:
                qs = qs.filter(**{
                    f'{field}__' + ('lt' if reverse else 'gt'): last_field,
                })
            return qs[0:items_per_round]
        return wrapper
    return inner


def to_snake_case(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    name = re.sub('__([A-Z])', r'_\1', name)
    name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name)
    return name.lower()


def to_camel_case(name):
    return ''.join(word.title() for word in name.split('_'))


def merge(a, b, path=None, raise_conflicts=True):
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            elif raise_conflicts:
                raise ValueError('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a
